#include "BStream.h"
#include "BOpcodeShell.h"
#include <stdio.h>


class HPoint 
{ 
public:
    float        x;	
    float        y;	
    float        z;	
    inline void Set(float X,float Y,float Z=0.0) { x=X,y=Y,z=Z; }

}; 

const float PI=3.1415926f;


static void CreateSphere(float *center, float radius, int numsides, 
    int *pcount_out, float **points_out, int *flistlen_out, int **faces_out )
{
    int i, j, htile, wtile, pts_cnt, flist_cnt;
    float theta, phi, dt, dp, x, y, z;
    HPoint *points;
    int *faces;

    htile = numsides/2;
    wtile = numsides;
    pts_cnt = htile*wtile;
    points = new HPoint[ pts_cnt ];
    // faces need wtile + 1 numbers to specify each sphere end cap
    // and 5*numsides numbers for each strip between the end caps.
    // There will be numsides/2-1 of these strips.
    flist_cnt = (wtile+1)*2 + 5*(htile-1)*wtile;
    faces = new int[ flist_cnt ];

    dt = 2.0f * PI / wtile;
    dp = 1.0f * PI / (htile+1);
    phi = dp;
    for( i = 0 ; i < htile ; i++ ) {
        theta = 0;
        for( j = 0 ; j < wtile ; j++ ) {
            x = radius * (float)cos(theta) * (float)sin(phi);
            y = radius * (float)cos(phi);
            z = radius * (float)sin(theta) * (float)sin(phi);
            points[ i*wtile + j ].Set( center[0] + x, center[1] + y, center[2] + z );
            theta += dt;
        }
        phi += dp;
    }

    // set the number of points in the end caps of the sphere in the connectivity list
    faces[5*(htile-1)*wtile] = wtile;
    faces[5*(htile-1)*wtile+wtile+1] = wtile;
    for (i = 0; i < (htile-1); i++)
    {
        for (j = 0; j < wtile; j++)
        {
            // generate the connectivity list for the side quads of the cylinder
            faces[(i*wtile + j)*5] = 4; // number of points in this polygon
            faces[(i*wtile + j)*5+1]=i*wtile + j;
            faces[(i*wtile + j)*5+2]=i*wtile + (j+1)%wtile; //next point along circle; wrap at numsides
            faces[(i*wtile + j)*5+3]=(i+1)*wtile + (j+1)%wtile;  // correcsponding point at top of cylinder
            faces[(i*wtile + j)*5+4]=(i+1)*wtile + j; //next point along circle; wrap at numsides
        }
    }
    for (j = 0; j < wtile; j++)
    {
        // generate the connectivity list for the ends of the cylinder
        faces[(htile-1)*wtile*5 + j + 1]= (wtile - j - 1);
        faces[(htile-1)*wtile*5 + j + 1 + wtile + 1]= wtile*(htile-1) + j;
    }
    *pcount_out = pts_cnt;
    *points_out = (float *)points;
    *flistlen_out = flist_cnt;
    *faces_out = (int *)faces;
}



// export the HSF object to a file

static void WriteObject(BBaseOpcodeHandler * handler, BStreamFileToolkit * tk, FILE * fp,
                        char * buffer, int size) 
{
    TK_Status status;

    // ask the toolkit to write out the object; the toolkit will write out as much as
    // the current buffer can hold
    status = handler->Write(*tk);

    // if there is still more data left to write; we have to loop through 
    // generating and writing buffers of data until we're done
    while (status == TK_Pending) 
    {
        // due to compression, preparing a new buffer might immediately fill that buffer
        // so we need an extra check to possibly export that buffer and prepare a new buffer
        while (status == TK_Pending) 
        {
            // ask the toolkit how much of the buffer is currently filled
            // this will normally be the same as the buffer size that we passed in
            int count = tk->CurrentBufferLength();

            // write the current out to our file since this example deals with 
            // exporting a file; however, instead of writing to a file, the buffer of
            // data could be sent to an application data-structure, or transmitted
            // over a network ('streamed'), etc...
            fwrite(buffer, sizeof (char), count, fp);

            // tell the toolkit where it can resume writing data; in this case, we
            // can reuse the same buffer once we're done with it
            status = tk->PrepareBuffer(buffer, size);
        }

        // ask the toolkit to continue trying to write out the object's data
        status = handler->Write(*tk);
    }
}


class TK_My_Shell: public TK_Shell
{
public:
    TK_My_Shell():TK_Shell() { };
    ~TK_My_Shell(){ Reset(); };

    TK_Status Interpret (BStreamFileToolkit & tk, long key, int lod) alter 
    {
        TK_Status status = TK_Normal;
        HPoint center;
        float radius = 1.0f;
        float bbox[6];
        float *points, *normals, *indices;
        int p_count, f_count, flist_length, *flist, i;
        bool is_firstpass;

        /* Take the write options in toolkit and from them compute an initial value for 
         * this opcode's suboption */
        is_firstpass = (lod==2);
        InitSubop( tk, lod, is_firstpass );
		/* Turn off default tristrip option if your database does not have good tristrips that you'd 
		 * like to use as the face list */
        SetSubop( ~TKSH_TRISTRIPS & GetSubop() );
        SetLodLevel( lod );
        SetKey( tk, key );

        if( key == 13)
            center.Set(0.0f,0.0f,0.0f);
        else
            center.Set(3.0f,0.0f,0.0f);

        // This is where the bounding box of the shell needs to be determined.  The 
        // bounding box is used for the local vertex compression.  In this example, we set it to the bounding
        // box of the sphere we're creating.
        bbox[0] = center.x - 1.0f;
        bbox[1] = center.y - 1.0f;
        bbox[2] = center.z - 1.0f;
        bbox[3] = center.x + 1.0f;
        bbox[4] = center.y + 1.0f;
        bbox[5] = center.z + 1.0f;
        SetBounding( bbox );

        // if the level=0, then we're exporting the original shell if we're spitting out the full resolution version
        if (GetLodLevel() == 0) {
            // This is where you will go out to your gfx database and get the data defining the shell.
            // We're using a simple sphere generator to create a point cloud and some connectivity
            CreateSphere((float *)&center, radius, 30, &p_count, &points, &flist_length, &flist);
            SetPoints(p_count, points);
            SetFaces(flist_length, flist);

            normals = new float[ 3 * p_count ];
            for ( i = 0 ; i < p_count * 3 ; i += 3) {
                normals[ i + 0 ] = points[ i + 0 ] - center.x;
                normals[ i + 1 ] = points[ i + 1 ] - center.y;
                normals[ i + 2 ] = points[ i + 2 ] - center.z;
            }
            SetVertexNormals( normals );
            delete [] normals;

            // now let's do the same for the face level coloring
            f_count = (flist_length - 2*30 - 2)/5 + 2;
            indices = new float[ f_count ];
            for ( i = 0 ; i < f_count ; i++ ) {
                indices[i] = (float)(i%4);
            }
            SetFaceIndices( indices );
            delete [] indices;
            delete [] points;
            delete [] flist;
        }
        else if (GetLodLevel() == 1) {
            // create a LOD version of the sphere where numsides=12
            CreateSphere((float *)&center, radius, 12, &p_count, &points, &flist_length, &flist);
            SetPoints(p_count, points);
            SetFaces(flist_length, flist);
            delete [] points;
            delete [] flist;
        }
        else {
            // LOD2 is a collection of non-shells consisting of exactly one circle
            TK_Circle *circhandler;

            circhandler = new TK_Circle( TKE_Circle );
            circhandler = (TK_Circle *)circhandler;
            circhandler->SetStart( center.x + radius, center.y, center.z );
            circhandler->SetMiddle( center.x - radius, center.y, center.z );
            circhandler->SetEnd( center.x, center.y + radius, center.z );
            AppendObject( tk, circhandler );
        }
        return status;
    }

    /* Here we clean up the mess we might have made with non-shell LODs. */
    void Reset() {
        BBaseOpcodeHandler *handler;
        while((handler = PopObject()) != NULL) {
            delete handler;
        }
        TK_Shell::Reset();
    }
};


int main() 
{
    // create a new file toolkit object
    BStreamFileToolkit *mytoolkit = new BStreamFileToolkit;

    char mybuffer[4096];
    int size=4096;
    int count;

    FILE *flptr;

    // open a binary file called 'testfile.hsf' for writing
    flptr=fopen("testfile3.hsf", "wb");

    // tell the toolkit where to begin writing data; the size of the 'utility' buffer
    // can be anything; for this example, we use a 4k buffer
    mytoolkit->PrepareBuffer(mybuffer, size);

    TK_Open_Segment *open_seg_handler; 
    TK_Close_Segment *close_seg_handler;
    TK_My_Shell *shell_handler;
    BBaseOpcodeHandler *handler;
    TK_Color_RGB *color_handler;
    TK_Matrix *matrix_handler;
    TK_Referenced_Segment *ref_seg_handler;
    TK_Camera *view_handler;

    // create a TK_Header object and export it's associated HSF objects,
    // which include the TKE_Comment and TKE_File_Info objects
    // TK_Header automatically formats the TKE_Comment to contain file version 
    // information.  
    handler = new TK_Header();
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // create a TK_Camera object to export a TKE_View opcode; this defines an initial
    // camera used to view the scene-graph contained in the file
    view_handler = new TK_Camera (TKE_View);
    //many apps have chosen to make the name "default" have special meaning -- that
    //it be used as the initial view.
    view_handler->SetView("default");
    view_handler->SetProjection(TKO_Camera_Perspective);
    view_handler->SetPosition(3.0, 0.0, -15.0);
    view_handler->SetTarget(3.0, 0.0, 0.0);
    view_handler->SetUpVector(1.0, 1.0, 0.0);
    view_handler->SetField(7.0, 7.0);
    WriteObject(view_handler, mytoolkit, flptr, mybuffer, size);
    delete view_handler;

    // export a TKE_Start_Compression object
    handler = new TK_Compression(TKE_Start_Compression);
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // Defining the colormap
    float colormap[]={1.0,1.0,1.0,0.5,0.5,0.5,1.0,0.0,0.0,0.0,0.0,1.0};
    handler = new TK_Color_Map;
    ((TK_Color_Map *)handler)->SetValues(4, colormap);
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // export the TKE_Open_Segment opcode for 'object_1' 
    open_seg_handler= new TK_Open_Segment;
    open_seg_handler->SetSegment("/include_library/object_1");
    WriteObject(open_seg_handler, mytoolkit, flptr, mybuffer, size);

    // We could delete this handler and recreate one the next time we need it, but since
    // we're going to reuse it, we instead call its Reset method and keep it around.
    // Our derived class would overload the Reset method and do cleanup/re-initialization
    // of any custom data.
    open_seg_handler->Reset();       

    // LOD 2 of the shell id'd by key 13
    // we're going to reuse this handler so Reset it and keep it around
    shell_handler = new TK_My_Shell;
    shell_handler->Interpret(*mytoolkit, 13, 2);
    WriteObject(shell_handler, mytoolkit, flptr, mybuffer, size);

    // Tags are actually automatically generated for shells. However, if you did want to tag 
    // something, call the Tag method of the BaseOpcodeHandler class.  For example:
    // handler->Tag(*mytoolkit, 2);

    // We could delete this handler and recreate one the next time we need it, but since
    // we're going to reuse it, we instead call its Reset method and keep it around.
    // Our derived class would overload the Reset method and do cleanup/re-initialization
    // of any custom data.
    shell_handler->Reset();    

    // we're going to reuse this handler so Reset it and keep it around
    close_seg_handler = new TK_Close_Segment;
    WriteObject(close_seg_handler, mytoolkit, flptr, mybuffer, size);
    close_seg_handler->Reset();

    // export the TKE_Open_Segment opcode for 'object_2' 
    open_seg_handler->SetSegment("/include_library/object_2");
    WriteObject(open_seg_handler, mytoolkit, flptr, mybuffer, size);
    open_seg_handler->Reset();

    // LOD 2 of the shell id'd by key 14
    shell_handler->Interpret(*mytoolkit, 14, 2);
    WriteObject(shell_handler, mytoolkit, flptr, mybuffer, size);
    shell_handler->Reset();

    WriteObject(close_seg_handler, mytoolkit, flptr, mybuffer, size);
    close_seg_handler->Reset();

    // export the TKE_Open_Segment opcode for 'part_1' 
    open_seg_handler->SetSegment("part_1");
    WriteObject(open_seg_handler, mytoolkit, flptr, mybuffer, size);
    open_seg_handler->Reset();

    // Review the Geometry_Bits enum in BOpcodeHandler.h if
    // you want to get more granularity on the color settings of different geometry in the segment
    color_handler = new TK_Color_RGB();
    color_handler->SetGeometry(1);
    color_handler->SetRGB(1.0, 0.0, 0.0);
    WriteObject(color_handler, mytoolkit, flptr, mybuffer, size);
    delete color_handler;

    // include the 'object_1' segment
    ref_seg_handler = new TK_Referenced_Segment(TKE_Include_Segment);
    ref_seg_handler->SetSegment("/include_library/object_1");
    WriteObject(ref_seg_handler, mytoolkit, flptr, mybuffer, size);
    ref_seg_handler->Reset();

    float mm1[] = {1.0, 0.0, 0.0, 0.0,
                   0.0, 1.0, 0.0, 0.0,
                   0.0, 0.0, 1.0, 0.0,
                   2.0, 0.0, 0.0, 1.0};

    matrix_handler = new TK_Matrix(TKE_Modelling_Matrix);
    matrix_handler->SetMatrix(mm1);
    WriteObject(matrix_handler, mytoolkit, flptr, mybuffer, size);
    matrix_handler->Reset();

    WriteObject(close_seg_handler, mytoolkit, flptr, mybuffer, size);
    close_seg_handler->Reset();


    // export the TKE_Open_Segment opcode for 'part_2' 
    open_seg_handler->SetSegment("part_2");
    WriteObject(open_seg_handler, mytoolkit, flptr, mybuffer, size);
    open_seg_handler->Reset();

    // include the 'object_2' segment
    ref_seg_handler->SetSegment("/include_library/object_2");
    WriteObject(ref_seg_handler, mytoolkit, flptr, mybuffer, size);
    ref_seg_handler->Reset();

    float mm2[] = {1.0, 0.0, 0.0, 0.0,
                   0.0, 1.0, 0.0, 0.0,
                   0.0, 0.0, 1.0, 0.0,
                   3.0, 0.0, 0.0, 1.0};

    matrix_handler->SetMatrix(mm2);
    WriteObject(matrix_handler, mytoolkit, flptr, mybuffer, size);
    matrix_handler->Reset();

    WriteObject(close_seg_handler, mytoolkit, flptr, mybuffer, size);
    close_seg_handler->Reset();

    //*********** LODS ***********
    // Notice how we're passing in the full resolution of the shell at a later point, yet the toolkit 
    // automatically associates it back to the LOD from the shell_key passed in.

    // LOD 1 of the shell id'd by key 13
    shell_handler->Interpret(*mytoolkit, 13, 1);
    WriteObject(shell_handler, mytoolkit, flptr, mybuffer, size);
    shell_handler->Reset();

    // LOD 1 of the shell id'd by key 14
    shell_handler->Interpret(*mytoolkit, 14, 1);
    WriteObject(shell_handler, mytoolkit, flptr, mybuffer, size);
    shell_handler->Reset();

    // Full resolution (LOD 0) of the shell id'd by key 13
    shell_handler->Interpret(*mytoolkit, 13, 0);
    WriteObject(shell_handler, mytoolkit, flptr, mybuffer, size);
    shell_handler->Reset();

    // Full resolution (LOD 0) of the shell id'd by key 14
    shell_handler->Interpret(*mytoolkit, 14, 0);
    WriteObject(shell_handler, mytoolkit, flptr, mybuffer, size);
    shell_handler->Reset();

    // export a TKE_Stop_Compression object
    handler = new TK_Compression(TKE_Stop_Compression);
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // export a TKE_Termination object
    handler = new TK_Terminator(TKE_Termination);
    WriteObject(handler, mytoolkit, flptr, mybuffer, size);
    delete handler;

    // ask the toolkit how much of the final buffer is filled
    count = mytoolkit->CurrentBufferLength();

    // write out the final buffer
    fwrite(mybuffer, sizeof (char), count, flptr);

    // close the file
    fclose(flptr);

    // clean up the opcode handlers that we dynamically created but left around for reuse
    delete open_seg_handler;
    delete close_seg_handler;
    delete shell_handler;
    delete ref_seg_handler;
    delete matrix_handler;

    // cleanup the toolkit object
    delete mytoolkit;

    return 0;
}

